package clogs;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.PrintWriter;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import clogs.parser.*;
import clogs.ast.*;
import clogs.codegen.*;
import clogs.staticanalysis.TypeCheck;
import clogs.staticanalysis.WellFormedCheck;
import clogs.transform.*;
import clogs.util.*;
import java_cup.runtime.*;


public class Driver
{
  public static void main (String[] args) 
    throws Exception
  {
    String phasesConfigFilename = "config/default.xml";
    if (args.length > 1) {
      phasesConfigFilename = args[1];
    }

    compile (args[0], "clogs-output.s", phasesConfigFilename);
  }


  public static void compile (String clogsSourceFilename, String clogsTargetFilename, String phasesConfigFilename)
    throws Exception
  {
    List<ExtDecl> edecls = parse (clogsSourceFilename);
    // print(edecls);
    List<FunDefSig> externals = getExternals ();
    List<Factory<Phase>> phases = loadPhases (phasesConfigFilename);
    edecls = runPhases (phases, edecls, externals);
    generateCode (edecls, clogsTargetFilename);
  }


  public static List<ExtDecl> parse (String clogsSourceFilename)
    throws Exception
  {
    SymbolFactory sf = new ComplexSymbolFactory ();
    ClogsLexer lexer = new ClogsLexer (new FileInputStream (clogsSourceFilename), sf);
    ClogsParser parser = new ClogsParser (lexer, sf);
    Symbol symbol = parser.parse ();
    List<ExtDecl> edecls = (List<ExtDecl>) symbol.value;
    return edecls;
  }


  static List<FunDefSig> getExternals ()
  {
    List<FunDefSig> externals = List.<FunDefSig>nil ();
    {
      List<Decl> decls = List.nil ();
      decls = decls.snoc (new Decl (new TypeArray (new TypeInt ()), "array", new Optional<Exp> ()));
      externals = externals.snoc (new FunDefSig (new TypeVoid (), "_print_array", decls));
    }
    {
      List<Decl> decls = List.nil ();
      decls = decls.snoc (new Decl (new TypeInt (), "n", new Optional<Exp> ()));
      externals = externals.snoc (new FunDefSig (new TypeVoid (), "_print_char", decls));
    }
    {
      List<Decl> decls = List.nil ();
      decls = decls.snoc (new Decl (new TypeInt (), "n", new Optional<Exp> ()));
      externals = externals.snoc (new FunDefSig (new TypeVoid (), "_print_int", decls));
    }
    {
      List<Decl> decls = List.nil ();
      externals = externals.snoc (new FunDefSig (new TypeVoid (), "_print_newline", decls));
    }

    return externals;
  }


  public static List<Factory<Phase>> loadPhases (String phasesConfigFilename) 
    throws Exception
  {
    String url = new File (phasesConfigFilename).getAbsoluteFile ().toURI ().toURL ().toString ();
    InputSource inputSource = new InputSource (url);
    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance ();
    dbf.setValidating (false);
    DocumentBuilder db = dbf.newDocumentBuilder ();
    Document document = db.parse (inputSource);

    List<Factory<Phase>> phases = List.nil ();
    NodeList nl = document.getElementsByTagName ("phase");
    for (int i = 0; i < nl.getLength (); i++) {
      Element phaseElt = (Element) nl.item (i);
      String className = phaseElt.getAttribute ("name");
      boolean printAfterTransform = Boolean.parseBoolean (phaseElt.getAttribute ("printAfterTransform"));
      boolean printAfterTransformAndTypeCheck = Boolean.parseBoolean (phaseElt.getAttribute ("printAfterTransformAndTypeCheck"));
      phases = phases.snoc (createPhaseFactory (className, printAfterTransform, printAfterTransformAndTypeCheck));
    }

    return phases;
  }


  static Factory<Phase> createPhaseFactory (final String className, final boolean printAfterTransform, final boolean printAfterTransformAndTypeCheck)
  {
    return new Factory<Phase> () 
      { 
        public Phase newInstance () 
        { 
          Phase p;
          try {
            Class<Phase> clazz = (Class<Phase>) Class.forName (className);
            p = clazz.newInstance ();
          } catch (Exception e) {
            throw new RuntimeException (e);
          }
          p.printAfterTransform = printAfterTransform; 
          p.printAfterTransformAndTypeCheck = printAfterTransformAndTypeCheck; 
          return p; 
        }
      };
  }


  static List<ExtDecl> runPhases (List<Factory<Phase>> phases, List<ExtDecl> edecls, List<FunDefSig> externals) 
  {
    for (Factory<Phase> factory : phases) {
      Phase p = factory.newInstance ();
      edecls = p.transform (edecls);
      if (p.printAfterTransform) {
        System.out.println ("RESULT OF: " + p.getClass ().getName ());
        print (edecls); separator ();
      }
      edecls = check (edecls, externals);
      if (p.printAfterTransformAndTypeCheck) {
        System.out.println ("TYPECHECKED RESULT OF: " + p.getClass ().getName ());
        print (edecls); separator ();
      }
    }
    return edecls;
  }


  static void separator ()
  {
    System.out.println ();
    System.out.println ("=======================================");
    System.out.println ();
  }


  static void print (List<ExtDecl> edecls)
  {
    //PrettyPrinter pp = new PrettyPrinter ();
    AnotherPrinter pp = new AnotherPrinter ();
    pp.printExtDecls (edecls);
    System.out.println (pp);
  }


  static List<ExtDecl> check (List<ExtDecl> edecls, List<FunDefSig> externals)
  {
    try {
      new WellFormedCheck (edecls, externals).check ();
      edecls = new TypeCheck (edecls, externals).check ();
    } catch (Exception e) {
      e.printStackTrace ();
      System.exit (1);
    }
    return edecls;
  }


  public static void generateCode (List<ExtDecl> edecls, String clogsTargetFilename)
    throws Exception
  {
    FileWriter fw = new FileWriter (clogsTargetFilename);
    PrintWriter pw = new PrintWriter (fw);
    new CodeGen (edecls, pw).generate ();
    pw.close ();
    fw.close ();
  }
}
