package clogs.ast;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Iterator;

import clogs.util.List;
import clogs.util.Optional;


public class AnotherPrinter
{
  private final StringWriter sw = new StringWriter ();
  private final PrintWriter pw = new PrintWriter (sw);
  private int indentation = 0;
  private boolean breakline = false;
  

  public String toString ()
  {
    return sw.toString ();
  }


  private void indent () 
  {
    indentation += 2; 
    breakline = true; 
  }


  private void outdent ()
  {
    indentation -= 2; 
    breakline = true; 
  }


  private void newline ()
  {
    breakline = true; 
  }


  public void print (Object obj) {
    if (breakline) {
      pw.println ("");
      for (int i = 0; i < indentation; i++) {
        pw.print (" "); 
      }
      breakline = false;
    }
    pw.print (obj);
  }


  public void printEmptyLine () {
    print ("");
    newline ();
  }


  public void printExtDecls (List<ExtDecl> edecls)
  {
    for (ExtDecl edecl : edecls) {
      printExtDecl (edecl);
      printEmptyLine ();
    }
  }


  public void printExtDecl (ExtDecl edecl)
  {
    if (edecl instanceof Decl) {
      Decl decl = (Decl) edecl;
      printDecl (decl);

    } else if (edecl instanceof FunDef) {
      FunDef fundef = (FunDef) edecl;
      printFunDef (fundef);

    } else {
      throw new RuntimeException ("Missing ExtDecl case: " + edecl.getClass ().getName ());
    }
  }


  public void printType (Type type)
  {
    if (type instanceof TypeArray) {
      TypeArray typeA = (TypeArray) type;
      print ("TypeArray (");
      printType (typeA.type);
      print (")");

    } else if (type instanceof TypeInt) {
      print ("TypeInt");

    } else if (type instanceof TypeVoid) {
      print ("TypeVoid");
    } else {
      throw new RuntimeException ("Missing Type case: " + type.getClass ().getName ());
    }
  }


  public void printParams (List<Decl> decls)
  {
    boolean first = true;
    print("List<Decl>[");
    for (Decl decl : decls) {
      if (!first) {
        print (", ");
      }
      first = false;
      printDecl (decl);
    }
    print("]");
  }


  public void printDecls (List<Decl> decls)
  {
    print ("List<Decl>[");
    newline ();
    indent ();
    for (Decl decl : decls) {
      printDecl (decl);
    }
    newline ();
    outdent ();
    print ("]");
    newline ();
  }


  public void printDecl (Decl decl)
  {
    print ("Decl (");
    printType (decl.type);
    print (", ");
    print (decl.name);
    if (!decl.eo.isEmpty ()) {
      print (", ");
      printExp (decl.eo.get ());
    }
    print (")");
  }


  public void printFunDef (FunDef fundef)
  {
    print ("FunDef (");
    printType (fundef.type);
    print (", ");
    print (fundef.name);
    print (", ");
    printParams (fundef.params);
    print (", ");
    indent ();
    newline ();
    printStat (fundef.body);
    outdent ();
    newline ();
    print (")");
  }


  public void printBlock (List<Stat> stats)
  {
    print ("List<Stat>[");
    newline ();
    indent ();
    for (Stat stat : stats) {
      printStat (stat);
    }
    newline ();
    outdent ();
    print ("]");
    newline ();
  }


  public void printStat (Stat stat)
  {
    for (String label : stat.labels) {
      print (label);
      print (":");
      newline ();
    }

    if (stat instanceof StatCompound) {
      StatCompound statC = (StatCompound) stat;
      print ("StatCompound (");
      indent ();
      newline ();
      printDecls (statC.decls);
      printBlock (statC.stats);
      outdent ();
      print (")");

    } else if (stat instanceof StatExp) {
      StatExp statE = (StatExp) stat;
      print ("StatExp (");
      printExp (statE.exp);
      print (")");

    } else if (stat instanceof StatGoto) {
      StatGoto statG = (StatGoto) stat;
      print ("StatGoto (");
      print (statG.target);
      print (")");

    } else if (stat instanceof StatIf) {
      StatIf statI = (StatIf) stat;
      print ("StatIf (");
      printExp (statI.exp);
      print (", ");
      indent ();
      newline ();
      printStat (statI.statT);
      outdent ();
      newline ();
      print (", ");
      indent ();
      newline ();
      printStat (statI.statF);
      outdent ();
      print (")");

    } else if (stat instanceof StatReturn) {
      StatReturn statR = (StatReturn) stat;
      print ("StatReturn (");
      if (!statR.oe.isEmpty ()) {
        printExp (statR.oe.get ());
      }
      print (")");

    } else if (stat instanceof StatSkip) {
      print ("StatSkip ()");
      
    } else if (stat instanceof StatWhile) {
      StatWhile statW = (StatWhile) stat;
      print ("StatWhile (");
      printExp (statW.exp);
      print (", ");
      indent ();
      newline ();
      printStat (statW.stat);
      outdent ();
      print (")");

    } else {
      throw new RuntimeException ("Missing Stat case: " + stat.getClass ().getName ());
    }

    newline ();
  }


  public void printExp (Exp exp)
  {
    if (!exp.to.isEmpty ()) {
      print ("(");
    }
    if (exp instanceof ExpArrayAccess) {
      ExpArrayAccess expA = (ExpArrayAccess) exp;
      print ("ExpArrayAccess (");
      printExp (expA.array);
      print (", ");
      printExp (expA.index);
      print (")");

    } else if (exp instanceof ExpAssign) {
      ExpAssign expA = (ExpAssign) exp;
      print ("ExpAssign (");
      printExp (expA.left);
      print (", ");
      printExp (expA.right);
      print (")");

    } else if (exp instanceof ExpBinOp) {
      ExpBinOp expB = (ExpBinOp) exp;
      print ("ExpBinOp (");
      printExp (expB.left);
      print (", ");
      print (expB.op);
      print (", ");
      printExp (expB.right);
      print (")");

    } else if (exp instanceof ExpComma) {
      ExpComma expC = (ExpComma) exp;
      print ("ExpComma (");
      printExp (expC.left);
      print (", ");
      printExp (expC.right);
      print (")");

    } else if (exp instanceof ExpFunCall) {
      ExpFunCall expF = (ExpFunCall) exp;
      print ("ExpFunCall (");
      print (expF.name);
      print (", ");
      boolean first = true;
      print ("List<Exp>[");
      for (Exp arg : expF.args) {
        if (!first) {
          print (", ");
        }
        first = false;
        printExp (arg);
      }
      print ("]");

    } else if (exp instanceof ExpInt) {
      ExpInt expI = (ExpInt) exp;
      print ("ExpInt (" + Integer.toString (expI.value) + ")");

    } else if (exp instanceof ExpNew) {
      ExpNew expN = (ExpNew) exp;
      print ("ExpNew (");
      print (expN.contentType);
      print (", ");
      print (expN.size);
      print (")");

    } else if (exp instanceof ExpString) {
      ExpString expS = (ExpString) exp;
      print ("ExpString (\"" + expS.value + "\")");

    } else if (exp instanceof ExpUnOp) {
      ExpUnOp expU = (ExpUnOp) exp;
      print ("ExpUnOp (");
      print (expU.op);
      print (", ");
      printExp (expU.exp);
      print (")");

    } else if (exp instanceof ExpVar) {
      ExpVar expV = (ExpVar) exp;
      print ("ExpVar (" + expV.name + ")");

    } else {
      throw new RuntimeException ("Missing Exp case: " + exp.getClass ().getName ());
    }

    if (!exp.to.isEmpty ()) {
      print (" :: ");
      printType (exp.to.get ());
      print (")");
    }

  }
}
