package clogs.codegen;

import java.io.PrintWriter;

import clogs.ast.*;
import clogs.transform.*;
import clogs.util.List;
import clogs.util.Map;
import clogs.util.Optional;
import clogs.staticanalysis.Context;


public class CodeGen
{
  private final List<ExtDecl> edecls;
  private final List<Decl> globals;
  private final List<FunDef> fundefs;
  private final DualPrintWriter dpw;
  public static final int intSize = 4; // an integer is stored in 4 bytes
  public static final int intSizeLog2 = 2; 


  public CodeGen (List<ExtDecl> edecls, PrintWriter pw)
  {
    this.edecls = edecls;

    List<Decl> globalsNew = List.nil ();
    List<FunDef> fundefsNew = List.nil ();
    for (ExtDecl edecl : edecls) {
      if (edecl instanceof Decl) {
        Decl decl = (Decl) edecl;
        globalsNew = globalsNew.snoc (decl);

      } else if (edecl instanceof FunDef) {
        FunDef fundef = (FunDef) edecl;
        fundefsNew = fundefsNew.snoc (fundef);

      } else {
        throw new RuntimeException ("Missing ExtDecl case: " + edecl.getClass ().getName ());
      }
    }
    this.globals = globalsNew;
    this.fundefs = fundefsNew;

    this.dpw = new DualPrintWriter (pw);
  }


  // Suppose that all the global variable declarations precede the
  // function declarations, then we have:
  // 
  //  code (T1 x1) = asmd1
  //  ...
  //  code (Tn xn) = asmdn
  //  context((decl1 ... decln fundef1 ... fundefm)) |- code (fundef1) = asmf1
  //  ...
  //  context((decl1 ... decln fundef1 ... fundefm)) |- code (fundefm) = asmfm
  //  ------------------------------------------------------------------------
  //  code ((decl1 ... decln fundef1 ... fundefm)) =
  //    .text
  //    asmf1
  //    ...
  //    asmfm
  //    .data
  //    asmd1
  //    ...
  //    asmdm
  //
  public void generate ()
  {
    Context context = Context.emptyContext;
    context = context.addGlobalVariables (globals);
    context = context.addFunDefs (fundefs);

    dpw.stage1.println (".text");
    dpw.stage2.println (".data");
    
    for (Decl decl : globals) {
      generateGlobal (decl);
    }

    for (FunDef fundef : fundefs) {
      generateFunDef (context, fundef);
    }
    
    dpw.close ();    
  }


  //  --------------------------------------------
  //  code ((T x;)) = 
  //    .align 4
  //    .global x
  //    x:
  //    .long  0
  //
  private void generateGlobal (Decl decl)
  {
    assert (decl.eo.isEmpty ());
    dpw.stage2.println ("\t.align 4");
    dpw.stage2.println (".global " + decl.name);
    dpw.stage2.println (decl.name + ":");
    dpw.stage2.println ("\t.long 0");
    newline (dpw.stage2);    
  }


  //  G,x1:S1,...,xn:Sn |- code (compoundStat) = asm
  //  -------------------------------------------------
  //  G |- code ((T f(S1 x1,...,Sn xn) compoundStat)) = 
  //        .align 4
  //        .global f
  //        f:
  //        pushl   %ebp
  //        movl    %esp, %ebp
  //        asm
  //
  private void generateFunDef (Context context, FunDef fundef)
  {
    comment (dpw.stage1, "Function: " + fundef.name);

    context = context.addFunParams (fundef);
    comment (dpw.stage1, "Context: " + context);

    dpw.stage1.println ("\t.align 4");
    dpw.stage1.println (".global " + fundef.name);
    dpw.stage1.println (fundef.name + ":");

    dpw.stage1.println ("\tpushl\t%ebp");
    dpw.stage1.println ("\tmovl\t%esp, %ebp");

    generateStatCompound (context, fundef.body);
  }


  // G,x1:T1,...,xn:Tn |- code (stat1) = asm1
  // ...
  // G,x1:T1,...,xn:Tn |- code (statm) = asmm
  // ---------------------------------------------------
  // G |- code ({ T1 x1; ... Tn xn; stat1 ... statm }) =
  //       subl $(n*intSize), %esp
  //       asm1
  //       ...
  //       asmm
  //       addl $(n*intSize), %esp
  //
  private void generateStatCompound (Context context, StatCompound statC)
  {
    comment (dpw.stage1, "Compound begins");

    context = context.addLocalVariables (statC.decls);
    comment (dpw.stage1, "Context: " + context);

    int size = 0;
    for (Decl decl : statC.decls) {
      assert (decl.eo.isEmpty ());
      size += intSize;
    }
    dpw.stage1.println ("\tsubl\t$" + size + ", %esp");

    for (Stat stat : statC.stats) {
      generateStat (context, stat);
    }

    dpw.stage1.println ("\taddl\t$" + size + ", %esp");

    comment (dpw.stage1, "Compound ends");
  }


  private void generateStat (Context context, Stat stat)
  {
    for (String label : stat.labels) {
      dpw.stage1.println (label + ":");
    }

    if (stat instanceof StatCompound) {
      // G,x1:T1,...,xn:Tn |- code (stat1) = asm1
      // ...
      // G,x1:T1,...,xn:Tn |- code (statm) = asmm
      // ---------------------------------------------------
      // G |- code ({ T1 x1; ... Tn xn; stat1 ... statm }) =
      //       subl $n %esp
      //       asm1
      //       ...
      //       asmm
      //       addl $n %esp
      //
      StatCompound statC = (StatCompound) stat;
      generateStatCompound (context, statC);

    } else if (stat instanceof StatExp) {
      // G |- code (exp) = asm
      // ---------------------------------------------------
      // G |- code ((exp;)) = asm
      //
      StatExp statE = (StatExp) stat;
      comment (dpw.stage1, stat.removeLabels ().toString ());
      generateExp (context, statE.exp);

    } else if (stat instanceof StatGoto) {
      // ---------------------------------------------------
      // G |- code ((goto l;)) = jmp l
      //
      StatGoto statG = (StatGoto) stat;
      comment (dpw.stage1, stat.removeLabels ().toString ());
      dpw.stage1.println ("\tjmp\t" + statG.target); 

    } else if (stat instanceof StatIf) {
      // G |- location (v) = v'
      // ---------------------------------------------
      // G |- code ((if (v) goto l1; else goto l2;)) = 
      //       movl   v', %eax
      //       testl  %eax, %eax
      //       jne    %l1
      //       jmp    %l2
      //
      StatIf statI = (StatIf) stat;
      comment (dpw.stage1, "if: " + statI.exp);
      assert (statI.statT instanceof StatGoto);
      assert (statI.statF instanceof StatGoto);
      String targetT = ((StatGoto) statI.statT).target;
      String targetF = ((StatGoto) statI.statF).target;
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, statI.exp) + ", %eax");
      dpw.stage1.println ("\ttestl\t%eax, %eax");
      dpw.stage1.println ("\tjne\t" + targetT);
      dpw.stage1.println ("\tjmp\t" + targetF);

    } else if (stat instanceof StatReturn) {
      // ---------------------------------------------
      // G |- code ((return;)) = 
      //       movl   %ebp, %esp
      //       popl   %ebp
      //       ret
      //
      // G |- location (v) = v'
      // ---------------------------------------------
      // G |- code ((return v;)) = 
      //       movl   v', %eax
      //       movl   %ebp, %esp
      //       popl   %ebp
      //       ret
      //
      StatReturn statR = (StatReturn) stat;
      comment (dpw.stage1, stat.removeLabels ().toString ());
      if (!statR.oe.isEmpty ()) {
        dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, statR.oe.get ()) + ", %eax");
      }
      dpw.stage1.println ("\tmovl\t%ebp, %esp");
      dpw.stage1.println ("\tpopl\t%ebp");
      dpw.stage1.println ("\tret");
      
    } else if (stat instanceof StatSkip) {
      // ---------------------------------------------
      // G |- code ((skip;)) = nop
      //
      StatSkip statS = (StatSkip) stat;
      comment (dpw.stage1, stat.removeLabels ().toString ());
      dpw.stage1.println ("\tnop\t");

    } else if (stat instanceof StatWhile) {
      StatWhile statW = (StatWhile) stat;
      assert (false);

    } else {
      throw new RuntimeException ("Missing Stat case: " + stat.getClass ().getName ());
    }
  }


  void generateExp (Context context, Exp exp) 
  {
    // We first translate the RHS, leaving the result in %eax.
    // Afterwards, move the value in %eax to the location indicated by
    // the LHS, if any.
    //
    Exp expLHS;
    Exp expRHS;
    if (exp instanceof ExpAssign) {
      ExpAssign expA = (ExpAssign) exp;
      expLHS = expA.left;
      expRHS = expA.right;
    } else {
      expLHS = null;
      expRHS = exp;
    }

    if (expRHS instanceof ExpArrayAccess) {
      // G |- location (x) = x'
      // G |- location (v) = v'
      // ---------------------------------------------
      // G |- code (x[v]) = 
      //       movl   x', %edx
      //       movl   v', %ecx
      //       sall   $(log_2(intSize)), %ecx
      //       addl   %ecx, %edx
      //       movl   0(%edx), %eax
      //
      ExpArrayAccess expA = (ExpArrayAccess) expRHS;
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expA.array) + ", %edx");
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expA.index) + ", %ecx");
      dpw.stage1.println ("\tsall\t$" + intSizeLog2 + ", %ecx"); 
      dpw.stage1.println ("\taddl\t%ecx, %edx");
      dpw.stage1.println ("\tmovl\t0(%edx), %eax");

    } else if (expRHS instanceof ExpAssign) {
      assert (false);

    } else if (expRHS instanceof ExpBinOp) {
      // G |- location (v1) = v1'
      // G |- location (v2) = v2'
      // ---------------------------------------------
      // G |- code ((v1 + v2)) = 
      //       movl   v1', %eax
      //       movl   v2', %ecx
      //       addl   %ecx, %eax
      //
      ExpBinOp expB = (ExpBinOp) expRHS;
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expB.left) + ", %eax");
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expB.right) + ", %ecx");
      if (expB.op.equals (ExpBinOp.BinOp.PLUS)) {
        dpw.stage1.println ("\taddl\t%ecx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.MINUS)) {
        dpw.stage1.println ("\tsubl\t%ecx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.TIMES)) {
        dpw.stage1.println ("\timull\t%ecx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.DIVIDE)) {
        // In Intel-syntax = cdq, sign-extend dword %eax to quad %edx:%eax 
        dpw.stage1.println ("\tcltd");
        // Divides %edx:%eax by %ecx with quotient in %eax and remainder in %edx
        dpw.stage1.println ("\tidivl\t%ecx");
      } else if (expB.op.equals (ExpBinOp.BinOp.MOD)) {
        // In Intel-syntax = cdq, sign-extend dword %eax to quad %edx:%eax 
        dpw.stage1.println ("\tcltd");
        // Divides %edx:%eax by %ecx with quotient in %eax and remainder in %edx
        dpw.stage1.println ("\tidivl\t%ecx");
        dpw.stage1.println ("\tmovl\t%edx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.LESSTHAN)) {
        dpw.stage1.println ("\tcmpl\t%ecx, %eax");
        dpw.stage1.println ("\tsetl\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.GREATERTHAN)) {
        dpw.stage1.println ("\tcmpl\t%ecx, %eax");
        dpw.stage1.println ("\tsetg\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.LEQ)) {
        dpw.stage1.println ("\tcmpl\t%ecx, %eax");
        dpw.stage1.println ("\tsetle\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.GEQ)) {
        dpw.stage1.println ("\tcmpl\t%ecx, %eax");
        dpw.stage1.println ("\tsetge\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.EQUALSEQUALS)) {
        dpw.stage1.println ("\tcmpl\t%ecx, %eax");
        dpw.stage1.println ("\tsete\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.BANGEQUALS)) {
        dpw.stage1.println ("\tcmpl\t%ecx, %eax");
        dpw.stage1.println ("\tsetne\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.AMP)) {
        dpw.stage1.println ("\tandl\t%ecx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.CARET)) {
        dpw.stage1.println ("\txorl\t%ecx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.BAR)) {
        dpw.stage1.println ("\torl\t%ecx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.AMPAMP)) {
        dpw.stage1.println ("\tcmpl\t$0, %eax");
        dpw.stage1.println ("\tsetne\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
        dpw.stage1.println ("\tcmpl\t$0, %ecx");
        dpw.stage1.println ("\tsetne\t%dl");
        dpw.stage1.println ("\tandl\t$1, %ecx");
        dpw.stage1.println ("\tandl\t%ecx, %eax");
      } else if (expB.op.equals (ExpBinOp.BinOp.BARBAR)) {
        dpw.stage1.println ("\tcmpl\t$0, %eax");
        dpw.stage1.println ("\tsetne\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
        dpw.stage1.println ("\tcmpl\t$0, %ecx");
        dpw.stage1.println ("\tsetne\t%dl");
        dpw.stage1.println ("\tandl\t$1, %ecx");
        dpw.stage1.println ("\torl\t%ecx, %eax");
      } else {
        assert (false);
      }

    } else if (expRHS instanceof ExpComma) {
      assert (false);

    } else if (expRHS instanceof ExpFunCall) {
      // G |- location (v1) = v1'
      // ...
      // G |- location (vn) = vn'
      // ---------------------------------------------
      // G |- code ((f(v1,...,vn))) = 
      //       pushl  vn'
      //       ...
      //       pushl  v1'
      //       call   f
      //       addl   $(n*intSize), %esp
      //
      ExpFunCall expF = (ExpFunCall) expRHS;
      for (Exp arg : expF.args.reverse ()) {
        dpw.stage1.println ("\tpushl\t" + getExpConstOrExpVar (context, arg));
      }
      dpw.stage1.println ("\tcall\t" + expF.name);
      dpw.stage1.println ("\taddl\t$" + (intSize * expF.args.size ()) + ", %esp");

    } else if (expRHS instanceof ExpInt) {
      // ---------------------------------------------
      // G |- code (n) = 
      //       movl   $n, %eax
      //
      ExpInt expI = (ExpInt) expRHS;
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expRHS) + ", %eax");

    } else if (expRHS instanceof ExpNew) {
      // G |- location (v) = v'
      // ---------------------------------------------
      // G |- code ((new T[v])) = 
      //       movl   v', %eax
      //       sall   $(log_2(intSize)), %eax
      //       push   %eax
      //       call   _malloc
      //       addl   $(1*intSize), %esp
      //
      ExpNew expN = (ExpNew) expRHS;
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expN.size) + ", %eax");
      dpw.stage1.println ("\tsall\t$" + intSizeLog2 + ", %eax"); 
      dpw.stage1.println ("\tpushl\t%eax");
      dpw.stage1.println ("\tcall\t_malloc");
      dpw.stage1.println ("\taddl\t$4, %esp");

    } else if (expRHS instanceof ExpString) {
      // ---------------------------------------------
      // G |- code (("value")) = 
      //       movl lab_string, %eax
      //       
      //       lab_string:
      //       .int  'v'
      //       .int  'a'
      //       .int  'l'
      //       .int  'u'
      //       .int  'e'
      //       .int  0
      //
      ExpString expS = (ExpString) expRHS;
      String label = Fresh.getLabel ("string_literal");
      dpw.stage2.println (label + ":\n");
      for (int i = 0; i < expS.value.length (); i++) {
        char ch = expS.value.charAt (i);
        dpw.stage2.println ("\t.int\t" + Integer.toString (0xffff & ch) + "\t/* \"" + expS.value.charAt (i) + "\" */");
      }
      dpw.stage2.println ("\t.int\t0");
      dpw.stage1.println ("\tmovl\t$" + label + ", %eax");

    } else if (expRHS instanceof ExpUnOp) {
      // G |- location (v) = v'
      // ---------------------------------------------
      // G |- code ((- v)) = 
      //       movl   v1', %eax
      //       negl   %eax
      //
      ExpUnOp expU = (ExpUnOp) expRHS;
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expU.exp) + ", %eax");
      if (expU.op.equals (ExpUnOp.UnOp.MINUS)) {
        dpw.stage1.println ("\tnegl\t%eax");
      } else if (expU.op.equals (ExpUnOp.UnOp.BANG)) {
        dpw.stage1.println ("\tcmpl\t$0, %eax");
        dpw.stage1.println ("\tsetne\t%al");
        dpw.stage1.println ("\tandl\t$1, %eax");
        dpw.stage1.println ("\txorl\t$1, %eax");
      } else {
        assert (false);
      }

    } else if (expRHS instanceof ExpVar) {
      // G |- location (x) = x'
      // ---------------------------------------------
      // G |- code (x) = 
      //       movl   x', %eax
      //
      ExpVar expV = (ExpVar) expRHS;
      dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expRHS) + ", %eax");

    } else {
      throw new RuntimeException ("Missing Exp case: " + exp.getClass ().getName ());
    }

    if (expLHS != null) {
      if (expLHS instanceof ExpArrayAccess) {
        // G |- location (x) = x'
        // G |- location (v) = v'
        // ---------------------------------------------
        // G |- code (x[v] = %eax) = 
        //       movl   x', %edx
        //       movl   v', %ecx
        //       sall   $(log_2(intSize)), %ecx
        //       addl   %ecx, %edx
        //       movl   %eax, 0(%edx)
        //
        ExpArrayAccess expA = (ExpArrayAccess) expLHS;
        dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expA.array) + ", %edx");
        dpw.stage1.println ("\tmovl\t" + getExpConstOrExpVar (context, expA.index) + ", %ecx");
        dpw.stage1.println ("\tsall\t$" + intSizeLog2 + ", %ecx"); 
        dpw.stage1.println ("\taddl\t%ecx, %edx");
        dpw.stage1.println ("\tmovl\t%eax, 0(%edx)");

      } else if (expLHS instanceof ExpVar) {
        // G |- location (x) = x'
        // ---------------------------------------------
        // G |- code (x = %eax) = 
        //       movl   %eax, x'
        //
        dpw.stage1.println ("\tmovl\t%eax, " + getExpConstOrExpVar (context, expLHS));

      } else {
        assert (false);
      }
    }
  }


  private String getExpConstOrExpVar (Context context, Exp exp)
  {
    if (exp instanceof ExpInt) {
      ExpInt expI = (ExpInt) exp;
      return "$" + expI.value;

    } else if (exp instanceof ExpVar) {
      ExpVar expV = (ExpVar) exp;
      return context.getVariableLocation (expV.name);

    } else {
      throw new RuntimeException ("Found non-variable expression: " + exp);
    }
  }


  private void comment (PrintWriter pw, String comment)
  {
    pw.println ("\t/* " + comment + " */");
  }


  private void newline (PrintWriter pw)
  {
    pw.println ();
  }


  private void separator (PrintWriter pw)
  {
    pw.println ();
    pw.println ("/* ############################################################ */");
    pw.println ();
  }
}

