package clogs.staticanalysis;

import java.util.NoSuchElementException;

import clogs.ast.*;
import clogs.codegen.CodeGen;
import clogs.util.List;
import clogs.util.Map;
import clogs.util.Optional;


public class Context
{
  protected final Map<String, Type> variableTypes;
  protected final Map<String, FunDefSig> fundefsigs;
  protected final Map<String, String> variableLocations;
  protected final int paramCounter;
  protected final int localVariableCounter;


  public static final Context emptyContext = 
    new Context (Map.<String, Type>empty (), Map.<String, FunDefSig>empty (), Map.<String, String>empty (), 2, -1);


  protected Context (Map<String, Type> variableTypes, Map<String, FunDefSig> fundefsigs, Map<String, String> variableLocations, 
                     int paramCounter, int localVariableCounter)
  {
    this.variableTypes = variableTypes;
    this.fundefsigs = fundefsigs;
    this.variableLocations = variableLocations;
    this.paramCounter = paramCounter;
    this.localVariableCounter = localVariableCounter;
  }


  public String getVariableLocation (String name)
  {
    String location = variableLocations.get (name);
    if (location == null) {
      throw new NoSuchElementException ("Variable " + name + " not found in context " + this);
    }
    return location;
  }
  

  public boolean containsVariable (String name)
  {
    Type type = variableTypes.get (name);
    return type != null;
  }


  public Type getVariableType (String name)
  {
    Type type = variableTypes.get (name);
    if (type == null) {
      throw new NoSuchElementException ("Variable " + name + " not found in context " + this);
    }
    return type;
  }
  

  public boolean containsFunDefSig (String name)
  {
    FunDefSig fundefsig = fundefsigs.get (name);
    return fundefsig != null;
  }


  public FunDefSig getFunDefSig (String name)
  {
    FunDefSig fundefsig = fundefsigs.get (name);
    if (fundefsig == null) {
      throw new NoSuchElementException ("Function definition signature " + name + " not found in context " + this);
    }
    return fundefsig;
  }


  public Context addGlobalVariables (List<Decl> decls)
  {
    Map<String, Type> variableTypesNew = variableTypes;
    Map<String, String> variableLocationsNew = variableLocations;
    for (Decl decl : decls) {
      if (variableTypesNew.get (decl.name) != null) {
        throw new RuntimeException ("Variable " + decl.name + " already found in context " + this);
      }
      variableTypesNew = variableTypesNew.add (decl.name, decl.type);
      variableLocationsNew = variableLocationsNew.add (decl.name, decl.name);
    }
    return new Context (variableTypesNew, fundefsigs, variableLocationsNew, paramCounter, localVariableCounter);
  }


  public Context addFunParams (FunDefSig fundefsig)
  {
    assert (paramCounter == 2);
    int paramCounterTmp = paramCounter;
    Map<String, Type> variableTypesNew = variableTypes;
    Map<String, String> variableLocationsNew = variableLocations;
    for (Decl decl : fundefsig.params) {
      if (variableTypesNew.get (decl.name) != null) {
        throw new RuntimeException ("Variable " + decl.name + " already found in context " + this);
      }
      variableTypesNew = variableTypesNew.add (decl.name, decl.type);
      variableLocationsNew = variableLocationsNew.add (decl.name, (CodeGen.intSize * paramCounterTmp++) + "(%ebp)");
    }
    return new Context (variableTypesNew, fundefsigs, variableLocationsNew, paramCounterTmp, localVariableCounter);
  }


  public Context addLocalVariables (List<Decl> decls)
  {
    int localVariableCounterTmp = localVariableCounter;
    Map<String, Type> variableTypesNew = variableTypes;
    Map<String, String> variableLocationsNew = variableLocations;
    for (Decl decl : decls) {
      if (variableTypesNew.get (decl.name) != null) {
        throw new RuntimeException ("Variable " + decl.name + " already found in context " + this);
      }
      variableTypesNew = variableTypesNew.add (decl.name, decl.type);
      variableLocationsNew = variableLocationsNew.add (decl.name, (CodeGen.intSize * localVariableCounterTmp--) + "(%ebp)");
    }
    return new Context (variableTypesNew, fundefsigs, variableLocationsNew, paramCounter, localVariableCounterTmp);
  }


  public Context addFunDefSig (FunDefSig fundefsig)
  {
    if (fundefsigs.get (fundefsig.name) != null) {
      throw new RuntimeException ("Function definition signature " + fundefsig.name + " already found in context " + this);
    }
    return new Context (variableTypes, fundefsigs.add (fundefsig.name, fundefsig), variableLocations, paramCounter, localVariableCounter);
  }


  public Context addFunDefSigs (List<FunDefSig> fundefsigs)
  {
    Context context = this;
    for (FunDefSig fundefsig : fundefsigs) {
      context = context.addFunDefSig (fundefsig);
    }
    return context;
  }


  public Context addFunDefs (List<FunDef> fundefs)
  {
    Context context = this;
    for (FunDef fundef : fundefs) {
      context = context.addFunDefSig (fundef);
    }
    return context;
  }


  public String toString ()
  {
    boolean first;
    StringBuffer sb = new StringBuffer ();
    first = true;
    sb.append ("variables: ");
    for (String var : variableTypes.keyList ()) {
      if (!first) {
        sb.append (", ");
      }
      first = false;
      Type type = variableTypes.get (var);
      String location = variableLocations.get (var);
      sb.append (var);
      sb.append (" :: ");
      sb.append (type);
      sb.append (" @ ");
      sb.append (location);
    }
    sb.append ("; ");
    first = true;
    for (String funName : fundefsigs.keyList ()) {
      if (!first) {
        sb.append (", ");
      }
      first = false;
      FunDefSig fundefsig = fundefsigs.get (funName);
      sb.append (fundefsig.type);
      sb.append (" "); 
      sb.append (fundefsig.name);
      sb.append (" ("); 
      sb.append (fundefsig.params);
      sb.append (")"); 
    }
    sb.append ("; paramCounter = ");
    sb.append (paramCounter);
    sb.append (")"); 
    return sb.toString ();
  }
}

