package clogs.staticanalysis;

import clogs.ast.*;
import clogs.util.List;
import clogs.util.Optional;


public class TypeCheck 
{
  public class TypeCheckException extends Exception
  {
    TypeCheckException (String s)
    {
      super (s);
    }
  }


  private final List<Decl> globalDecls;
  private final List<FunDef> fundefsOnly;
  private final List<FunDefSig> fundefsigs;


  public TypeCheck (List<ExtDecl> edecls, List<FunDefSig> fundefsigs)
  {
    List<Decl> declsTmp = List.nil ();
    List<FunDef> fundefsOnlyTmp = List.nil ();
    List<FunDefSig> fundefsigsTmp = fundefsigs;
    for (ExtDecl edecl : edecls) {
      if (edecl instanceof Decl) {
        Decl decl = (Decl) edecl;
        declsTmp = declsTmp.snoc (decl);

      } else if (edecl instanceof FunDef) {
        FunDef fundef = (FunDef) edecl;
        fundefsOnlyTmp = fundefsOnlyTmp.snoc (fundef);
        fundefsigsTmp = fundefsigsTmp.snoc (fundef);

      } else {
        throw new RuntimeException ("Missing ExtDecl case: " + edecl.getClass ().getName ());
      }
    }
    
    this.globalDecls = declsTmp;
    this.fundefsOnly = fundefsOnlyTmp;
    this.fundefsigs = fundefsigsTmp;
  }


  // In comments, let S and T range over types and G over contexts.
  // Contexts have the form:
  //
  // G ::= empty
  //     | G,x:T                (variable declaration)
  //     | G,f:(S1,...,Sn)->T   (function declaration)
  //
  // The judgments are written:
  //
  //   |- extdecl1 ... extdecln : Program
  // G |- extdecl : Declaration
  // G |- stat : Statement<declaredReturnType>
  // G |- exp : T
  //   |- exp is lvalue
  //


  // To check a bunch of global variable and function declarations, we
  // put them in the context, then start checking them; this allows
  // mutually recursive definitions.
  //
  // context(extdecl1 ... extdecln) |- extdecl1 : Declaration
  // ...
  // context(extdecl1 ... extdecln) |- extdecln : Declaration
  // ------------------------------------------------------------------
  //                                |- extdecl1 ... extdecln : Program
  //
  public List<ExtDecl> check ()
    throws TypeCheckException
  {
    Context context = Context.emptyContext;

    context = context.addGlobalVariables (globalDecls);
    context = context.addFunDefSigs (fundefsigs);

    List<Decl> declsNew = checkDecls (context, globalDecls);
    List<FunDef> fundefsOnlyNew = checkFunDefSigs (context);

    List<ExtDecl> result = List.nil ();
    for (Decl decl : declsNew) {
      result = result.snoc (decl);
    }
    for (FunDef fundef : fundefsOnlyNew) {
      result = result.snoc (fundef);
    }

    return result;
  }


  // To check a variable declaration, we just need to make sure that
  // the initializer has the correct type:
  //
  // ------------------------------
  // G |- (T x;) : Declaration
  //
  // G |- exp : T
  // -------------------------------
  // G |- (T x = exp;) : Declaration
  // 
  List<Decl> checkDecls (Context context, List<Decl> decls)
    throws TypeCheckException
  {
    List<Decl> result = List.nil ();
    for (Decl decl : decls) {
      if (decl.eo.isEmpty ()) {
        result = result.snoc (decl);
      } else {
        Exp exp = checkExp (context, decl.eo.get ());
        if (!matchesDeclaredType (decl.type, exp)) {
          throw new TypeCheckException ("Declared type " + decl.type + " does not match computed type for " + exp);
        }
        result = result.snoc (new Decl (decl.type, decl.name, new Optional<Exp> (exp)));
      }
    }
    return result;
  }

  
  // To check a function declaration, we check the body:
  //
  // G,x1:S1,...,xn:Sn |- compoundStat : T
  // -------------------------------------------------------
  // G |- (T f (S1 x1,...,Sn xn) compoundStat) : Declaration
  //
  List<FunDef> checkFunDefSigs (Context context)
    throws TypeCheckException
  {
    List<FunDef> fundefsOnlyNew = List.nil ();
    for (FunDef fundef : fundefsOnly) {
      StatCompound bodyNew = checkStatCompound (context.addFunParams (fundef), fundef.type, fundef.body);
      FunDef fundefNew = new FunDef (fundef.type, fundef.name, fundef.params, bodyNew);
      fundefsOnlyNew = fundefsOnlyNew.snoc (fundefNew);
    }
    return fundefsOnlyNew;
  }


  // To check a compound statement, check each statement and initializer
  // G,x1:T1,...,xn:Tn |- (T1 x1 = exp1;) : Declaration
  // ...
  // G,x1:T1,...,xn:Tn |- (Tn xn = expn;) : Declaration
  // G,x1:T1,...,xn:Tn |- stat1 : Statement<T>
  // ...
  // G,x1:T1,...,xn:Tn |- statm : Statement<T>
  // -----------------------------------------------------------------------
  // G |- { T1 x1 = exp1; ... Tn xn = expn; stat1 ... statm } : Statement<T>
  //
  StatCompound checkStatCompound (Context context, Type declaredReturnType, StatCompound statC)
    throws TypeCheckException
  {
    context = context.addLocalVariables (statC.decls);
    List<Decl> declsNew = checkDecls (context, statC.decls);
    List<Stat> statsNew = List.nil ();
    for (Stat stat : statC.stats) {
      Stat statNew = checkStat (context, declaredReturnType, stat);
      statsNew = statsNew.snoc (statNew);
    }
    return new StatCompound (declsNew, statsNew);
  }


  // To check a statment, do a case analysis.
  Stat checkStat (Context context, Type declaredReturnType, Stat stat)
    throws TypeCheckException
  {
    if (stat instanceof StatCompound) {
      // G,decls |- stats : Statement<T>
      // ------------------------------------
      // G |- { decls stats } : Statement<T>
      //
      StatCompound statC = (StatCompound) stat;
      return checkStatCompound (context, declaredReturnType, statC).addLabels (stat.labels);

    } else if (stat instanceof StatExp) {
      // G |- exp : S    (for some type S)
      // ---------------------------------------
      // G |- (exp;) : Statement<T>
      //
      StatExp statE = (StatExp) stat;
      return new StatExp (checkExp (context, statE.exp)).addLabels (stat.labels);

    } else if (stat instanceof StatGoto) {
      // -----------------------------------
      // G |- (goto l;) : Statement<T>
      //
      return stat;

    } else if (stat instanceof StatIf) {
      // G |- exp : int
      // G |- statT : Statement<T>
      // G |- statF : Statement<T>
      // -----------------------------------------------------
      // G |- (if (exp) statT else statF) : Statement<T>
      //
      StatIf statI = (StatIf) stat;
      Exp expNew = checkExp (context, statI.exp);
      if (!matchesDeclaredType (new TypeInt (), expNew)) {
        throw new TypeCheckException ("Expression in conditional has non-integer type " + expNew);
      }
      return new StatIf (expNew, checkStat (context, declaredReturnType, statI.statT), checkStat (context, declaredReturnType, statI.statF)).addLabels (stat.labels);

    } else if (stat instanceof StatReturn) {
      // ---------------------------------------
      // G |- (return;) : Statement<void>
      //
      // G |- exp : T
      // ---------------------------------------
      // G |- (return exp;) : Statement<T>
      //
      StatReturn statR = (StatReturn) stat;
      if (statR.oe.isEmpty ()) {
        if (!testTypeEquality (declaredReturnType, new TypeVoid ())) {
          throw new TypeCheckException ("Empty return for function with non-void type");
        }
        return statR;
      } else {
        Exp expNew = checkExp (context, statR.oe.get ());
        if (!matchesDeclaredType (declaredReturnType, expNew)) {
          throw new TypeCheckException ("Declared return type " + declaredReturnType + " does not match computed type for " + expNew);
        }
        return new StatReturn (new Optional<Exp> (expNew)).addLabels (stat.labels);
      }

    } else if (stat instanceof StatSkip) {
      // ---------------------------------
      // G |- (skip;) : Statement<T>
      return stat;

    } else if (stat instanceof StatWhile) {
      // G |- exp : int
      // G |- statW : Statement<T>
      // -----------------------------------------------------
      // G |- (while (exp) statW) : Statement<T>
      //
      StatWhile statW = (StatWhile) stat;
      Exp expNew = checkExp (context, statW.exp);
      if (!matchesDeclaredType (new TypeInt (), expNew)) {
        throw new TypeCheckException ("Expression in while has non-integer type " + expNew);
      }
      return new StatWhile (expNew, checkStat (context, declaredReturnType, statW.stat)).addLabels (stat.labels);

    } else {
      throw new RuntimeException ("Missing Stat case: " + stat.getClass ().getName ());
    }
  }


  // To check an expression, do a case analysis.
  Exp checkExp (Context context, Exp exp)
    throws TypeCheckException
  {
    Exp result;

    if (exp instanceof ExpArrayAccess) {
      // G |- array : T[]
      // G |- index : int
      // -----------------------------
      // G |- (array[index]) : T
      //
      ExpArrayAccess expA = (ExpArrayAccess) exp;
      Exp exp1 = checkExp (context, expA.array);
      Exp exp2 = checkExp (context, expA.index);
      assert (!exp1.to.isEmpty ());
      Type type1 = exp1.to.get ();
      if (!(type1 instanceof TypeArray)) {
        throw new TypeCheckException ("Array expression has non-array type " + exp1);
      }
      Type typeResult = ((TypeArray) type1).type;
      if (!matchesDeclaredType (new TypeInt (), exp2)) {
        throw new TypeCheckException ("Array index expression has non-integer type " + exp2);
      }
      result = new ExpArrayAccess (exp1, exp2).setType (typeResult);

    } else if (exp instanceof ExpAssign) {
      // G |- left : T
      // G |- right : T
      //   |- left is lval
      // ------------------------------
      // G |- (left = right) : T
      //
      ExpAssign expA = (ExpAssign) exp;
      Exp exp1 = checkExp (context, expA.left);
      Exp exp2 = checkExp (context, expA.right);
      assert (!exp1.to.isEmpty ());
      assert (!exp2.to.isEmpty ());
      if (!testTypeEquality (exp1.to.get (), exp2.to.get ())) {
        throw new TypeCheckException ("Unequal types in assignment from " + exp2 + " to " + exp1);
      }
      if (!isLValue (exp1)) {
        throw new TypeCheckException ("Assignment to non-lvalue " + exp1);
      }
      result = new ExpAssign (exp1, exp2).setType (exp2.to.get ());

    } else if (exp instanceof ExpBinOp) {
      // G |- left : int
      // G |- right : int
      // -----------------------------------
      // G |- (left binop right) : int
      //
      ExpBinOp expB = (ExpBinOp) exp;
      Exp exp1 = checkExp (context, expB.left);
      Exp exp2 = checkExp (context, expB.right);
      // At present, all binary operators have type "int * int ->
      // int", so we do not need to perform a case analysis on the
      // operator.
      if (!matchesDeclaredType (new TypeInt (), exp1)) {
        throw new TypeCheckException ("Argument to " + expB.op + " has non-integer type " + exp1);
      }
      if (!matchesDeclaredType (new TypeInt (), exp2)) {
        throw new TypeCheckException ("Argument to " + expB.op + " has non-integer type " + exp2);
      }
      result = new ExpBinOp (expB.op, exp1, exp2).setType (new TypeInt ());

    } else if (exp instanceof ExpComma) {
      // G |- left : S
      // G |- right : T
      // -----------------------------
      // G |- (left , right) : T
      //
      ExpComma expC = (ExpComma) exp;
      Exp exp1 = checkExp (context, expC.left);
      Exp exp2 = checkExp (context, expC.right);
      assert (!exp2.to.isEmpty ());
      result = new ExpComma (exp1, exp2).setType (exp2.to.get ());

    } else if (exp instanceof ExpFunCall) {
      // G(name) = (S1,...,Sn)->T
      // G |- arg1 : S1
      // ...
      // G |- argn : Sn
      // ------------------------------------
      // G |- (name(arg1,...,argn)) : T
      //
      ExpFunCall expF = (ExpFunCall) exp;
      if (!context.containsFunDefSig (expF.name)) {
        throw new TypeCheckException ("Function " + expF.name + " not found in context " + context);
      }
      FunDefSig fundefsig = context.getFunDefSig (expF.name);
      if (fundefsig.params.size () != expF.args.size ()) {
        throw new TypeCheckException ("Function " + expF.name + " requires " + fundefsig.params.size () + " arguments, " + 
                                      "but " + expF.args.size () + " arguments were given in " + expF);
      }
      List<Decl> decls = fundefsig.params;
      List<Exp> argsNew = List.nil ();
      for (Exp arg : expF.args) {
        Exp argNew = checkExp (context, arg);
        if (!matchesDeclaredType (decls.head ().type, argNew)) {
          throw new TypeCheckException ("Function " + expF.name + " parameter type " + decls.head () + " does not match computed type for " + arg);
        }
        argsNew = argsNew.snoc (argNew);
        decls = decls.tail ();
      }
      result = new ExpFunCall (expF.name, argsNew).setType (fundefsig.type);

    } else if (exp instanceof ExpInt) {
      // ------------
      // G |- n : int
      //
      ExpInt expI = (ExpInt) exp;
      result = expI.setType (new TypeInt ());

    } else if (exp instanceof ExpNew) {
      // G |- size : int
      // --------------------------------------------
      // G |- (new contentType[size]) : contentType[]
      //
      ExpNew expN = (ExpNew) exp;
      Exp exp1 = checkExp (context, expN.size);
      if (!matchesDeclaredType (new TypeInt (), exp1)) {
        throw new TypeCheckException ("Argument to new " + expN + " has non-integer type " + exp1);
      }
      result = new ExpNew (expN.contentType, exp1).setType (new TypeArray (expN.contentType));

    } else if (exp instanceof ExpString) {
      // --------------------
      // G |- "value" : int[]
      //
      ExpString expS = (ExpString) exp;
      result = expS.setType (new TypeArray (new TypeInt ()));

    } else if (exp instanceof ExpUnOp) {
      // G |- exp : int
      // ---------------------
      // G |- (unop exp) : int
      //
      ExpUnOp expU = (ExpUnOp) exp;
      Exp exp1 = checkExp (context, expU.exp);
      // At present, all unary operators have type "int -> int", so we
      // do not need to perform a case analysis on the operator.
      if (!matchesDeclaredType (new TypeInt (), exp1)) {
        throw new TypeCheckException ("Argument to " + expU.op + " has non-integer type " + exp1);
      }
      result = new ExpUnOp (expU.op, exp1).setType (new TypeInt ());

    } else if (exp instanceof ExpVar) {
      // G(x) = T
      // ----------
      // G |- x : T
      //
      ExpVar expV = (ExpVar) exp;
      if (!context.containsVariable (expV.name)) {
        throw new TypeCheckException ("Variable " + expV.name + " not found in context " + context);
      } 
      result = expV.setType (context.getVariableType (expV.name));

    } else {
      throw new RuntimeException ("Missing Exp case: " + exp.getClass ().getName ());
    }

    if (!exp.to.isEmpty () && !matchesDeclaredType (exp.to.get (), result)) {
      throw new TypeCheckException ("Declared type for expression " + exp + " does not match computed type for expression " + result);
    }
    
    return result;
  }


  boolean matchesDeclaredType (Type declared, Exp exp)
  {
    assert (!exp.to.isEmpty ());
    Type computed = exp.to.get ();
    return testTypeEquality (declared, computed);
  }

  
  boolean testTypeEquality (Type t1, Type t2) 
  {
    if ((t1 instanceof TypeVoid) && (t2 instanceof TypeVoid)) {
      return true;
    } else if ((t1 instanceof TypeInt) && (t2 instanceof TypeInt)) {
      return true;
    } else if ((t1 instanceof TypeArray) && (t2 instanceof TypeArray)) {
      TypeArray t1A = (TypeArray) t1;
      TypeArray t2A = (TypeArray) t2;
      return testTypeEquality (t1A.type, t2A.type);
    } else {
      return false;
    }
  }


  // Every variable and array access is an lvalue.
  // Nothing else is an lvalue.
  boolean isLValue (Exp exp)
  {
    if (exp instanceof ExpArrayAccess) {
      // -------------------------
      // |- (array[index]) is lvalue
      //
      return true;
    } else if (exp instanceof ExpVar) {
      // --------------
      // |- x is lvalue
      //
      return true;
    } else {
      return false;
    }
  }
}
