package clogs.staticanalysis;

import clogs.ast.*;
import clogs.util.List;
import clogs.util.Optional;


// Well-formedness addresses the following issues:
// 1. No void array types
// 2. All parameter and variable declarations have non-void types
// 3. No two global variable declarations have the same name
// 4. No two functions declarations have the same name
// 5. No two labels in the same function declaration have the same name
// 6. Every label used in a function is declared in the same function
// 7. Top level declarations have no initializers
// 8. Only top-level compound statements have declarations
public class WellFormedCheck 
{
  public class WellFormedCheckException extends Exception
  {
    WellFormedCheckException (String s)
    {
      super (s);
    }
  }


  private final List<Decl> globalDecls;
  private final List<FunDefSig> fundefsigs;


  public WellFormedCheck (List<ExtDecl> edecls, List<FunDefSig> fundefsigs)
  {
    List<Decl> declsTmp = List.nil ();
    List<FunDefSig> fundefsigsTmp = fundefsigs;
    for (ExtDecl edecl : edecls) {
      if (edecl instanceof Decl) {
        Decl decl = (Decl) edecl;
        declsTmp = declsTmp.snoc (decl);

      } else if (edecl instanceof FunDef) {
        FunDef fundef = (FunDef) edecl;
        fundefsigsTmp = fundefsigsTmp.snoc (fundef);

      } else {
        throw new RuntimeException ("Missing ExtDecl case: " + edecl.getClass ().getName ());
      }
    }
    
    this.globalDecls = declsTmp;
    this.fundefsigs = fundefsigsTmp;
  }


  public void check ()
    throws WellFormedCheckException
  {
    checkDecls (globalDecls, false);
    checkFunDefSigs ();
  }


  void checkDecls (List<Decl> decls, boolean initializerOK)
    throws WellFormedCheckException
  {
    // Ensure no declarations share the same name.
    checkNoDuplicateDecl (decls);

    for (Decl decl : decls) {
      // Ensure type is well formed.
      if (!checkTypeWellFormed (decl.type)) {
        throw new RuntimeException ("Type must be well formed: " + decl);
      }

      // Ensure type is suitable for a variable declaration.
      if (!checkNonReturnTypeOK (decl.type)) {
        throw new RuntimeException ("Variable type must be non-void: " + decl);
      }

      // Ensure no initializers in certain locations, e.g., global variables and function parameters
      if (!initializerOK && !decl.eo.isEmpty ()) {
        throw new RuntimeException ("Initializer not allowed here: " + decl);
      }
    }
  }


  void checkFunDefSigs ()
    throws WellFormedCheckException
  {
    // Ensure no function declarations share the same name.
    checkNoDuplicateFunDefSigs ();

    for (FunDefSig fundefsig : fundefsigs) {
      // Ensure type is well formed.
      if (!checkTypeWellFormed (fundefsig.type)) {
        throw new RuntimeException ("Type must be well formed: " + fundefsig.type);
      }

      checkDecls (fundefsig.params, false);

      if (fundefsig instanceof FunDef) {
        FunDef fundef = (FunDef) fundefsig;
        checkStatCompound (fundef.body, true);
      }
    }
  }


  void checkStatCompound (StatCompound statC, boolean topLevel)
    throws WellFormedCheckException
  {
    if (!statC.decls.isNil() && !topLevel) {
      throw new WellFormedCheckException ("Declaration not at top level in \"" + statC + "\"");
    }
    checkDecls (statC.decls, true);
      
    
    // Gather labels defined and used in compound statement.
    GatherLabelsInStat glisc = new GatherLabelsInStat ();
    glisc.gather (statC);

    // Ensure no duplicate label definitions.
    String duplicate = glisc.labelsDefined.findDuplicate ();
    if (duplicate != null) {
      throw new WellFormedCheckException ("Found duplicate label \"" + duplicate + "\" definition in \"" + statC + "\"");
    }

    // Ensure all used labels are defined in the same top-level compound statement.
    if (topLevel) {
      for (String label : glisc.labelsUsed) {
        if (!glisc.labelsDefined.contains (label)) {
          throw new WellFormedCheckException ("Label \"" + label + "\" used but not defined in \"" + statC + "\"");
        }
      }
    }
    
    for (Stat stat : statC.stats) {
      checkStat (stat);
    }
  }


  void checkStat (Stat stat)
    throws WellFormedCheckException
  {
    if (stat instanceof StatCompound) {
      StatCompound statC = (StatCompound) stat;
      checkStatCompound (statC, false);

    } else if (stat instanceof StatExp) {
      // Nothing to check. 

    } else if (stat instanceof StatGoto) {
      StatGoto statG = (StatGoto) stat;
      // Nothing to check. 

    } else if (stat instanceof StatIf) {
      StatIf statI = (StatIf) stat;
      checkStat (statI.statT);
      checkStat (statI.statF);

    } else if (stat instanceof StatReturn) {
      StatReturn statR = (StatReturn) stat;
      // Nothing to check. 

    } else if (stat instanceof StatSkip) {
      StatSkip statS = (StatSkip) stat;
      // Nothing to check. 

    } else if (stat instanceof StatWhile) {
      StatWhile statW = (StatWhile) stat;
      checkStat (statW.stat);

    } else {
      throw new RuntimeException ("Missing Stat case: " + stat.getClass ().getName ());
    }
  }


  void checkNoDuplicateDecl (List<Decl> decls)
    throws WellFormedCheckException
  {
    List<String> names = List.nil ();
    for (Decl decl : decls) {
      names = names.cons (decl.name);
    }

    String duplicate = names.findDuplicate ();
    if (duplicate != null) {
      throw new WellFormedCheckException ("Found duplicate declaration name for \"" + duplicate + "\"");
    }
  }


  void checkNoDuplicateFunDefSigs ()
    throws WellFormedCheckException
  {
    List<String> names = List.nil ();
    for (FunDefSig fundefsig : fundefsigs) {
      names = names.cons (fundefsig.name);
    }

    String duplicate = names.findDuplicate ();
    if (duplicate != null) {
      throw new WellFormedCheckException ("Found duplicate function definition declaration/signature for \"" + duplicate + "\"");
    }
  }


  boolean checkNonReturnTypeOK (Type type)
  {
    if (type instanceof TypeVoid) {
      return false;
    } else {
      return true;
    }
  }


  boolean checkTypeWellFormed (Type type)
  {
    if ((type instanceof TypeInt) || (type instanceof TypeVoid)) {
      return true;
    } 

    while (type instanceof TypeArray) {
      TypeArray typeA = (TypeArray) type;
      type = typeA.type;
    }

    if (type instanceof TypeInt) {
      return true;
    } else if (type instanceof TypeVoid) {
      return false;
    } else {
      throw new RuntimeException ("Missing Type case: " + type.getClass ().getName ());
    }
  }
}


