package clogs.transform;

import clogs.ast.*;
import clogs.util.List;
import clogs.util.Optional;


public class IntroduceThreeAddressCode extends Phase
{
  public List<ExtDecl> transform (List<ExtDecl> edecls)
  {
    List<ExtDecl> result = List.nil ();

    for (ExtDecl edecl : edecls) {
      if (edecl instanceof Decl) {
        Decl decl = (Decl) edecl;
        assert (decl.eo.isEmpty ());
        result = result.snoc (decl);

      } else if (edecl instanceof FunDef) {
        FunDef fundef = (FunDef) edecl;
        result = result.snoc (new FunDef (fundef.type, fundef.name, fundef.params, transformBody (fundef.body)));

      } else {
        throw new RuntimeException ("Missing ExtDecl case: " + edecl.getClass ().getName ());
      }
    }

    return result;
  }

  static class DeclsStats
  {
    final List<Decl> decls;
    final List<Stat> stats;

    DeclsStats (List<Decl> decls, List<Stat> stats)
    {
      this.decls = decls;
      this.stats = stats;
    }
  }


  static class DeclsStatsVariable
  {
    final List<Decl> decls;
    final List<Stat> stats;
    final Optional<Exp> oe; // A call to a function with void return type does not produce a result.

    DeclsStatsVariable (List<Decl> decls, List<Stat> stats, Optional<Exp> oe)
    {
      assert ((oe.isEmpty ()) || (oe.get () instanceof ExpInt) || (oe.get () instanceof ExpVar));
      this.decls = decls;
      this.stats = stats;
      this.oe = oe;
    }
  }


  // Write compound statements as
  //   { decls stats }
  //
  // Write instances of DeclsStats as
  //   < decls | stats > 
  //
  // Write instances of DeclsStatsVariable as
  //   < decls | stats | Optional<v> >
  // If the exp must be present, then:
  //   < decls | stats | v >
  //
  // The relevant functions are:
  //   transformBody (({ decls stats })) = { decls' stats' }
  //   transformStat (stat) = < decls | stats >
  //   transformExp (exp) = < decls | stats | Optional<v> >
  //

  
  static StatCompound transformBody (StatCompound statC) {
    // transformStat (stat_0) = < decls_0 | stats_0 >
    // ...
    // transformStat (stat_n) = < decls_n | stats_n >
    // -----------------------------------------------------
    // transformBody (({decls stat_0 ... stat_n}))
    //   = { decls decls_0 ... decls_n stats_0 ... stats_m }
    //
    List<Decl> declsNew = List.nil ();
    List<Stat> statsNew = List.nil ();
    for (Stat stat : statC.stats) {
      DeclsStats ds = transformStat (stat);
      declsNew = declsNew.concat (ds.decls);
      statsNew = statsNew.concat (ds.stats);
    }
    return new StatCompound (statC.decls.concat (declsNew), statsNew);
  }


  static DeclsStats transformStat (Stat stat) {
    List<Decl> declsNew = List.nil ();
    List<Stat> statsNew = List.nil ();

    if (stat instanceof StatCompound) {
      // transformBody (statC) = statC'
      // -----------------------------------------------------
      // transformStat (statC) = < nil | statC' >
      //
      StatCompound statC = (StatCompound) stat;
      for (Stat s : statC.stats) {
        DeclsStats ds = transformStat (s);
        declsNew = declsNew.concat (ds.decls);
        statsNew = statsNew.concat (ds.stats);
      }
      statsNew = List.singleton ((Stat)new StatCompound (statC.decls, statsNew));
      statsNew = addLabelsToHead (statsNew, statC.labels);

    } else if (stat instanceof StatExp) {
      // transformExp (exp) = < decls | stats | ... >
      // -----------------------------------------------------
      // transformStat ((exp;)) = < decls | stats >
      //
      StatExp statE = (StatExp) stat;
      DeclsStatsVariable sav = transformExp (statE.exp);
      declsNew = declsNew.concat (sav.decls);
      statsNew = statsNew.concat (sav.stats); 
      if (!statE.labels.isNil()) {
        if (statsNew.isNil()) {
          statsNew = statsNew.snoc (new StatSkip());
        }
        statsNew = addLabelsToHead (statsNew, statE.labels);
      }
      
    } else if (stat instanceof StatGoto) {
      // transformStat ((goto exp;)) = < nil | (goto exp;) >
      //
      statsNew = statsNew.snoc (stat);

    } else if (stat instanceof StatIf) {
      // transformExp (exp) = < decls | stats | v >
      // -----------------------------------------------------
      // transformStat ((if (exp) goto l1; else goto l2;))
      //   = < decls | stats  (if (v) goto l1; else goto l2;) >
      //
      StatIf statI = (StatIf) stat;
      assert (statI.statT instanceof StatGoto);
      assert (statI.statF instanceof StatGoto);
      DeclsStatsVariable sav = transformExp (statI.exp);
      declsNew = declsNew.concat (sav.decls);
      statsNew = statsNew.concat (sav.stats); 
      statsNew = statsNew.snoc (new StatIf (sav.oe.get (), statI.statT, statI.statF));
      statsNew = addLabelsToHead (statsNew, statI.labels);

    } else if (stat instanceof StatReturn) {
      // -----------------------------------------------------
      // transformStat ((return;)) = < nil | (return;) >
      //
      // transformExp (exp) = < decls | stats | v >
      // -----------------------------------------------------------------
      // transformStat ((return exp;)) = < decls | stats (return v;) >
      //
      StatReturn statR = (StatReturn) stat;
      if (statR.oe.isEmpty ()) {
        statsNew = statsNew.snoc (stat);
      } else {
        DeclsStatsVariable sav = transformExp (statR.oe.get ());
        declsNew = declsNew.concat (sav.decls);
        statsNew = statsNew.concat (sav.stats); 
        statsNew = statsNew.snoc (new StatReturn (new Optional<Exp> (sav.oe.get ())));
        statsNew = addLabelsToHead (statsNew, statR.labels);
      }

    } else if (stat instanceof StatSkip) {
      // transformStat ((skip;)) = < nil | (skip;) >
      //
      statsNew = statsNew.snoc (stat);

    } else if (stat instanceof StatWhile) {
      assert (false);

    } else {
      throw new RuntimeException ("Missing Stat case: " + stat.getClass ().getName ());
    }

    return new DeclsStats (declsNew, statsNew);
  }


  static List<Stat> addLabelsToHead (List<Stat> stats, List<String> labels)
  {
    Stat headNew = stats.head ().addLabels (labels);
    return stats.tail ().cons (headNew);
  }

  
  static DeclsStatsVariable transformExp (Exp exp) 
  {
    if (exp instanceof ExpArrayAccess) {
      // transformExp (array) = < decls1 | stats1 | v1 >
      // transformExp (index) = < decls2 | stats2 | v2 >
      // ------------------------------------------------------------------------------
      // transformExp ((array[index]) :: resultType)
      //   = < decls1 decls2 (resultType x;) | stats1 stats2 (x = (v1[v2]);) | x >
      //
      ExpArrayAccess expA = (ExpArrayAccess) exp;
      DeclsStatsVariable sav1 = transformExp (expA.array);
      DeclsStatsVariable sav2 = transformExp (expA.index);
      List<Decl> declsNew = sav1.decls.concat (sav2.decls);
      List<Stat> statsNew = sav1.stats.concat (sav2.stats);
      String labelNew = Fresh.getVar () + "_array";
      Exp expNew = new ExpVar (labelNew);
      declsNew = declsNew.snoc (new Decl (expA.to.get (), labelNew, new Optional<Exp> ()));
      statsNew = statsNew.snoc (new StatExp (new ExpAssign (expNew, new ExpArrayAccess (sav1.oe.get (), sav2.oe.get ()))));
      return new DeclsStatsVariable (declsNew, statsNew, new Optional<Exp> (expNew));

    } else if (exp instanceof ExpAssign) {
      // transformExp (right) = < decls3 | stats3 | v3 >
      // -----------------------------------------------------
      // transformExp ((x = right))
      //   = < decls3 | stats3 (x = v3;) | v3 >
      //
      // transformExp (array) = < decls1 | stats1 | v1 >
      // transformExp (index) = < decls2 | stats2 | v2 >
      // transformExp (right) = < decls3 | stats3 | v3 >
      // -------------------------------------------------------------------------------
      // transformExp ((array[index] = right))
      //   = < decls1 decls2 decls3 | stats1 stats2 stats3 (v1[v2] = v3;) | v3 >
      //
      ExpAssign expA = (ExpAssign) exp;
      List<Decl> declsNew = List.nil ();
      List<Stat> statsNew = List.nil ();
      Exp left;
      if (expA.left instanceof ExpArrayAccess) {
        ExpArrayAccess expAA = (ExpArrayAccess) expA.left;
        DeclsStatsVariable sav1 = transformExp (expAA.array);
        declsNew = declsNew.concat (sav1.decls);
        statsNew = statsNew.concat (sav1.stats);
        DeclsStatsVariable sav2 = transformExp (expAA.index);
        declsNew = declsNew.concat (sav2.decls);
        statsNew = statsNew.concat (sav2.stats);
        left = new ExpArrayAccess (sav1.oe.get (), sav2.oe.get ());
      } else if (expA.left instanceof ExpVar) {
        left = expA.left;
      } else {
        assert (false); // Impossible to reach this case in a well-typed program.
        left = null;
      } 
      DeclsStatsVariable sav3 = transformExp (expA.right);
      declsNew = declsNew.concat (sav3.decls);
      statsNew = statsNew.concat (sav3.stats);
      statsNew = statsNew.snoc (new StatExp (new ExpAssign (left, sav3.oe.get ())));
      return new DeclsStatsVariable (declsNew, statsNew, sav3.oe);

    } else if (exp instanceof ExpBinOp) {
      // transformExp (left)  = < decls1 | stats1 | v1 >
      // transformExp (right) = < decls2 | stats2 | v2 >
      // ----------------------------------------------------------------------------------
      // transformExp ((left op right) :: resultType)
      //   = < decls1 decls2 (resultType x;) | stats1 stats2 (x = (v1 op v2);) | x >
      //
      ExpBinOp expB = (ExpBinOp) exp;
      DeclsStatsVariable sav1 = transformExp (expB.left);
      DeclsStatsVariable sav2 = transformExp (expB.right);
      List<Decl> declsNew = sav1.decls.concat (sav2.decls);
      List<Stat> statsNew = sav1.stats.concat (sav2.stats);
      String labelNew = Fresh.getVar () + "_binop";
      Exp expNew = new ExpVar (labelNew);
      declsNew = declsNew.snoc (new Decl (expB.to.get (), labelNew, new Optional<Exp> ()));
      statsNew = statsNew.snoc (new StatExp (new ExpAssign (expNew, new ExpBinOp (expB.op, sav1.oe.get (), sav2.oe.get ()))));
      return new DeclsStatsVariable (declsNew, statsNew, new Optional<Exp> (expNew));

    } else if (exp instanceof ExpComma) {
      // transformExp (left)  = < decls1 | stats1 | Optional<exp1> >
      // transformExp (right) = < decls2 | stats2 | Optional<exp2> >
      // -----------------------------------------------------------
      // transformExp ((left , right))
      //   = < decls1 decls2 | stats1 stats2 | Optional<exp2> >
      //
      ExpComma expC = (ExpComma) exp;
      DeclsStatsVariable sav1 = transformExp (expC.left);
      DeclsStatsVariable sav2 = transformExp (expC.right);
      List<Decl> declsNew = sav1.decls.concat (sav2.decls);
      List<Stat> statsNew = sav1.stats.concat (sav2.stats);
      return new DeclsStatsVariable (declsNew, statsNew, sav2.oe);

    } else if (exp instanceof ExpFunCall) {
      // transformExp (arg1) = < decls1 | stats1 | v1 >
      // ...
      // transformExp (argn) = < declsn | statsn | vn >
      // -------------------------------------------------------------------------
      // transformExp ((name(arg1,...,argn)) :: void)
      //   = < decls1 ... declsn | stats1 ... statsn (name(v1,...,vn)) | nil >
      // AND
      // transformExp ((name(arg1,...,argn)) :: resultType)
      //   = < decls1 ... declsn (resultType x;) | stats1 ... statsn (x = (name(v1,...,vn))) | x >
      //
      ExpFunCall expF = (ExpFunCall) exp;
      List<Decl> declsNew = List.nil ();
      List<Stat> statsNew = List.nil ();
      List<Exp> argsNew = List.nil ();
      for (Exp exp2 : expF.args) {
        DeclsStatsVariable sav1 = transformExp (exp2);
        declsNew = declsNew.concat (sav1.decls);
        statsNew = statsNew.concat (sav1.stats);
        argsNew = argsNew.snoc (sav1.oe.get ());
      }
      if (expF.to.get () instanceof TypeVoid) {
        statsNew = statsNew.snoc (new StatExp (new ExpFunCall (expF.name, argsNew)));
        return new DeclsStatsVariable (declsNew, statsNew, new Optional<Exp> ());
      } else {
        String labelNew = Fresh.getVar () + "_funcall";
        Exp expNew = new ExpVar (labelNew);
        declsNew = declsNew.snoc (new Decl (expF.to.get (), labelNew, new Optional<Exp> ()));
        statsNew = statsNew.snoc (new StatExp (new ExpAssign (expNew, new ExpFunCall (expF.name, argsNew))));
        return new DeclsStatsVariable (declsNew, statsNew, new Optional<Exp> (expNew));
      }

    } else if (exp instanceof ExpInt) {
      // -----------------------------------------------------
      // transformExp (n) = < nil | nil | n >
      //
      return new DeclsStatsVariable (List.<Decl>nil (), List.<Stat>nil (), new Optional<Exp> (exp));

    } else if (exp instanceof ExpNew) {
      // transformExp (size) = < decls | stats | v >
      // -----------------------------------------------------
      // transformExp ((new contentType[size];))
      //   = < decls (contentType[] x;) | stats (x = (new contentType[v]);) | x >
      //
      ExpNew expN = (ExpNew) exp;
      DeclsStatsVariable sav = transformExp (expN.size);
      List<Decl> declsNew = sav.decls;
      List<Stat> statsNew = sav.stats;
      String labelNew = Fresh.getVar () + "_new";
      Exp expNew = new ExpVar (labelNew);
      declsNew = declsNew.snoc (new Decl (new TypeArray (expN.contentType), labelNew, new Optional<Exp> ()));
      statsNew = statsNew.snoc (new StatExp (new ExpAssign (expNew, new ExpNew (expN.contentType, sav.oe.get ()))));
      return new DeclsStatsVariable (declsNew, statsNew, new Optional<Exp> (expNew));

    } else if (exp instanceof ExpString) {
      // -----------------------------------------------------
      // transformExp (("value"))
      //   = < (int[] x;) | (x = ("value");) | x >
      //
      ExpString expS = (ExpString) exp;
      String labelNew = Fresh.getVar () + "_string_literal";
      Exp expNew = new ExpVar (labelNew);
      List<Decl> declsNew = List.nil ();
      List<Stat> statsNew = List.nil ();
      declsNew = declsNew.snoc (new Decl (new TypeArray (new TypeInt ()), labelNew, new Optional<Exp> ()));
      statsNew = statsNew.snoc (new StatExp (new ExpAssign (expNew, new ExpString (expS.value))));
      return new DeclsStatsVariable (declsNew, statsNew, new Optional<Exp> (expNew));

    } else if (exp instanceof ExpUnOp) {
      // transformExp (exp) = < decls | stats | v >
      // ------------------------------------------------------------
      // transformExp ((unop exp) :: resultType)
      //   = < decls (resultType x;) | stats (x = (unop v);) | x >
      //
      ExpUnOp expU = (ExpUnOp) exp;
      DeclsStatsVariable sav = transformExp (expU.exp);
      List<Decl> declsNew = sav.decls;
      List<Stat> statsNew = sav.stats;
      String labelNew = Fresh.getVar () + "_unop";
      Exp expNew = new ExpVar (labelNew);
      declsNew = declsNew.snoc (new Decl (expU.to.get (), labelNew, new Optional<Exp> ()));
      statsNew = statsNew.snoc (new StatExp (new ExpAssign (expNew, new ExpUnOp (expU.op, sav.oe.get ()))));
      return new DeclsStatsVariable (declsNew, statsNew, new Optional<Exp> (expNew));

    } else if (exp instanceof ExpVar) {
      // -----------------------------------------------------
      // transformExp (x) = < nil | nil | x >
      //
      return new DeclsStatsVariable (List.<Decl>nil (), List.<Stat>nil (), new Optional<Exp> (exp));

    } else {
      throw new RuntimeException ("Missing Exp case: " + exp.getClass ().getName ());
    }
  }
}
