package clogs.util;

import java.util.HashSet;
import java.util.Iterator;
import java.util.NoSuchElementException;


public abstract class List<A> implements Iterable<A>
{
  public static <A> List<A> nil ()
  {
    return new ListNil<A> ();
  }


  public static <A> List<A> singleton (A head)
  {
    return List.<A>nil ().cons (head);
  }


  public List<A> cons (A head)
  {
    return new ListCons<A> (head, this);
  }


  public List<A> snoc (A last)
  {
    return this.concat (singleton (last));
  }


  public A get (int index) 
  {
    int i = index;
    List<A> l = this;
    while ((i > 0) && (!l.isNil ())) {
      i--;
      l = l.tail ();
    }

    if ((i != 0) || l.isNil ()) {
      throw new IndexOutOfBoundsException 
        ("Attempt to access element " + i + " in list of length " + size () + "\n" + this);
    }

    return l.head ();
  }


  public abstract boolean isNil ();
  public abstract int size ();

  public abstract A head () throws EmptyListException;
  public abstract List<A> tail () throws EmptyListException;

  public abstract List<A> concat (List<A> l);
  public abstract List<A> reverse ();
  abstract List<A> reverseAux (List<A> end);


  public String toString ()
  {
    StringBuffer sb = new StringBuffer ();

    if (isNil ()) {
      return ("[]");
    }

    sb.append ("[ ");
    List l = this;
    boolean first = true;
    while (!l.isNil ()) {
      if (first) {
        first = false;
      } else {
        sb.append (", ");
      }
      sb.append (l.head ());
      l = l.tail ();
    }
    sb.append (" ]");
    
    return sb.toString ();
  }
  
  
  // Returns a duplicate in the list (upto .equals ()) if there is one, otherwise returns null.
  public A findDuplicate ()
  {
    HashSet<A> set = new HashSet<A> ();
    List<A> list = this;
    while (!list.isNil ()) {
      A o = list.head ();
      if (!set.contains (o)) {
        set.add (o);
      } else {
        return o;
      }
      list = list.tail ();
    }
    return null;
  }
  
  
  class ListIteratorImpl<A> implements Iterator<A> 
  {
    private List<A> contents;
    
    ListIteratorImpl (List<A> contents)
    {
      this.contents = contents;
    }
    
    public boolean hasNext ()
    {
      return !contents.isNil ();
    }
    
    public A next ()
    {
      if (contents.isNil ()) {
        throw new NoSuchElementException ();
      } else {
        A result = contents.head ();
        contents = contents.tail ();
        return result;
      }
    }
    
    public void remove ()
    {
      throw new UnsupportedOperationException ();
    }
  }
  
  
  public Iterator<A> iterator ()
  {
    return new ListIteratorImpl<A> (this);
  }
  
  
  // Applies fun.f to every object in the list to create a new list of the results.
  // If fun.f returns null, the result is discarded (a combined map/filter).
  public <B> List<B> map (Function<A,B> fun)
  {
    if (isNil ()) {
      return nil ();
    } else {
      B newHead = fun.f (head ());
      List<B> newTail = tail ().map (fun);
      if (newHead == null) {
        return newTail;
      } else {
        return newTail.cons (newHead);
      }
    }
  }


  public boolean contains (A o)
  {
    List<A> list = this;
    while (!list.isNil ()) {
      if (o.equals (list.head ())) {
        return true;
      }
      list = list.tail ();
    }
    return false;
  }
}


